#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <math.h>
#include <float.h>

#ifndef TILE_SIZE
#define TILE_SIZE 16
#endif

// Device function for GELU activation
__device__ inline float gelu_activation(float x) {
    return 0.5f * x * (1.0f + erff(x * 0.70710678f));
}

//--------------------------------------------------------------------------
// Modular device functions for tiled matrix multiplication with shared memory
//--------------------------------------------------------------------------

// Loads a tile of matrix A into shared memory
__device__ inline void load_A_tile(const float* __restrict__ A, float* A_tile, int row, int t, int M, int K) {
    int col = t * TILE_SIZE + threadIdx.x;
    if (row < M && col < K)
        A_tile[threadIdx.y * TILE_SIZE + threadIdx.x] = A[row * K + col];
    else
        A_tile[threadIdx.y * TILE_SIZE + threadIdx.x] = 0.0f;
}

// Loads a tile of matrix B into shared memory
// Note: B is stored row-major, but we access it as if it were transposed
__device__ inline void load_B_tile(const float* __restrict__ B, float* B_tile, int col, int t, int K, int N) {
    int rowB = t * TILE_SIZE + threadIdx.y;
    if (col < N && rowB < K)
        B_tile[threadIdx.y * TILE_SIZE + threadIdx.x] = B[col * K + rowB];
    else
        B_tile[threadIdx.y * TILE_SIZE + threadIdx.x] = 0.0f;
}

// Computes the product of the loaded A and B tiles
__device__ inline float compute_tile_product(const float* A_tile, const float* B_tile) {
    float sum = 0.0f;
    #pragma unroll
    for (int i = 0; i < TILE_SIZE; i++) {
        sum += A_tile[threadIdx.y * TILE_SIZE + i] * B_tile[i * TILE_SIZE + threadIdx.x];
    }
    return sum;
}

//--------------------------------------------------------------------------
// Modular Fused Matrix Multiplication with Bias Addition Kernel
// Computes: C = A * (B^T) + bias
// M: number of rows in A
// N: number of rows in B (output features)
// K: common dimension
//--------------------------------------------------------------------------
__global__ void ModularFusedMatMulBiasKernel(const float* __restrict__ A,
                                              const float* __restrict__ B,
                                              const float* __restrict__ bias,
                                              float* __restrict__ C,
                                              int M, int N, int K) {
    // Shared memory tiles for A and B
    __shared__ float A_tile[TILE_SIZE * TILE_SIZE];
    __shared__ float B_tile[TILE_SIZE * TILE_SIZE];

    int row = blockIdx.y * TILE_SIZE + threadIdx.y;
    int col = blockIdx.x * TILE_SIZE + threadIdx.x;
    float sum = 0.0f;
    
    int numTiles = (K + TILE_SIZE - 1) / TILE_SIZE;
    for (int t = 0; t < numTiles; t++) {
        load_A_tile(A, A_tile, row, t, M, K);
        load_B_tile(B, B_tile, col, t, K, N);
        __syncthreads();  // ensure the tiles are loaded
        sum += compute_tile_product(A_tile, B_tile);
        __syncthreads();  // prepare for next tile
    }
    if (row < M && col < N) {
        C[row * N + col] = sum + bias[col];
    }
}

//--------------------------------------------------------------------------
// Modular device functions for pooling, activation, scaling and reduction
//--------------------------------------------------------------------------

// Computes the average pooling over a pooling window, applies GELU and scales the result
__device__ inline float compute_pool_activation(const float* __restrict__ row_ptr, int N, int start, int pool_kernel_size, float scale_factor) {
    float sum = 0.0f;
    int count = 0;
    #pragma unroll
    for (int j = 0; j < pool_kernel_size; j++) {
        int col = start + j;
        if (col < N) {
            sum += row_ptr[col];
            count++;
        }
    }
    float avg = (count > 0) ? (sum / count) : 0.0f;
    return gelu_activation(avg) * scale_factor;
}

// Warp-level maximum reduction using shuffle operations
__device__ inline float warpReduceMax(float val) {
    for (int offset = warpSize / 2; offset > 0; offset /= 2) {
        val = fmaxf(val, __shfl_down_sync(0xffffffff, val, offset));
    }
    return val;
}

//--------------------------------------------------------------------------
// Modular Fused Pooling, Activation, Scaling and Max Reduction Kernel
// Processes one row per block
//--------------------------------------------------------------------------
__global__ void ModularFusedPoolActMaxKernel(const float* __restrict__ linear_output,
                                              float* __restrict__ output,
                                              int N,
                                              int pool_kernel_size,
                                              int output_length,
                                              float scale_factor) {
    int row = blockIdx.x; // one block per row
    int tid = threadIdx.x;
    float local_max = -FLT_MAX;
    const float* row_ptr = linear_output + row * N;

    // Each thread processes multiple pooling bins in a strided loop
    for (int bin = tid; bin < output_length; bin += blockDim.x) {
        int start = bin * pool_kernel_size;
        float activated = compute_pool_activation(row_ptr, N, start, pool_kernel_size, scale_factor);
        local_max = fmaxf(local_max, activated);
    }

    // Warp-level reduction
    local_max = warpReduceMax(local_max);

    // Use shared memory to store per-warp results
    __shared__ float shared_max[32]; // up to 32 warps per block
    int lane = threadIdx.x % warpSize;
    int warpId = threadIdx.x / warpSize;
    if (lane == 0) {
        shared_max[warpId] = local_max;
    }
    __syncthreads();

    // First warp reduces the warp results
    if (warpId == 0) {
        local_max = (tid < (blockDim.x + warpSize - 1) / warpSize) ? shared_max[lane] : -FLT_MAX;
        local_max = warpReduceMax(local_max);
        if (lane == 0) {
            output[row] = local_max;
        }
    }
}

//--------------------------------------------------------------------------
// Forward function: Chains the modular operations
// 1. Computes linear transformation using ModularFusedMatMulBiasKernel
// 2. Applies fused pooling, GELU activation, scaling, and max reduction
//--------------------------------------------------------------------------

torch::Tensor forward(
    torch::Tensor x,
    int pool_kernel_size,
    float scale_factor,
    torch::Tensor weight,
    torch::Tensor bias) {

    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");

    // Ensure tensors are contiguous
    x = x.contiguous();
    weight = weight.contiguous();
    bias = bias.contiguous();

    int M = x.size(0);   // Batch size
    int K = x.size(1);   // Input features
    int N = weight.size(0); // Output features

    auto options = torch::TensorOptions().dtype(x.dtype()).device(x.device());
    auto linear_output = torch::empty({M, N}, options);

    // Launch the modular fused matrix multiplication with bias addition kernel
    dim3 blockDimMat(TILE_SIZE, TILE_SIZE);
    dim3 gridDimMat((N + TILE_SIZE - 1) / TILE_SIZE, (M + TILE_SIZE - 1) / TILE_SIZE);
    ModularFusedMatMulBiasKernel<<<gridDimMat, blockDimMat>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        linear_output.data_ptr<float>(),
        M, N, K);

    // Compute the number of pooling bins
    int output_length = (N + pool_kernel_size - 1) / pool_kernel_size;
    auto output = torch::empty({M}, options);

    // Launch the modular fused pooling, activation, scaling and max reduction kernel
    int threads = 256;
    ModularFusedPoolActMaxKernel<<<M, threads>>>(
        linear_output.data_ptr<float>(),
        output.data_ptr<float>(),
        N,
        pool_kernel_size,
        output_length,
        scale_factor
    );

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Modular Fused CUDA forward (MatMul+Bias, Pool, GELU, Scale, Max Reduction)");
}
